from __future__ import print_function
import os, logging, shelve, math
from itertools import combinations
from qgis.core import QgsVectorLayer, QgsCoordinateReferenceSystem, QgsFeatureRequest
from PyQt4.QtCore import QPyNullVariant
from PyQt4.QtGui import QMessageBox

DICT_IDS = {"Epuntuali": "ID_ep","Ind_pu": "ID_SPU","Ind_ln": "ID_SLN","Geoidr": "ID_gi","Elineari": "ID_el","Forme": "ID_f","Instab_geotec": "ID_i","Geotec": "ID_gt","Stab": "ID_z","Instab": "ID_i"}

class StructureValidator(object):

	def __init__(self, path, version=4, _format="shp"):
		super(StructureValidator, self).__init__()

		self.__path = path
		self.__version = version
		self.__format = "shapefile" if _format == "shp" else "geodatabase"
		self.__dataStructure = self.get_data_structure()

	@property
	def path(self):
		return self.__path

	def setPath(self, path):
		self.__path = path

	@property
	def version(self):
		return self.__version

	@property
	def _format(self):
		return self.__format

	@property
	def dataStructure(self):
		return self.__dataStructure

	def get_data_structure(self):
		rootDir = os.path.join(os.getcwd(), "template_data")
		#rootDir = os.path.join(os.path.realpath(__file__), "template_data")
		for single in os.listdir(rootDir):
			print(single)
			if os.path.isdir(os.path.join(rootDir,single)):
				ver, form = single.split("_")
				print(ver, form)
				if ver == "v%d" % self.version and form == self._format:
					dataStructure = []
					for root, dirs, files in os.walk(os.path.join(rootDir, single)):
						if root.lower().endswith("cle"):
							continue
						for singleFile in files:
							name = singleFile.lower()
							if name.endswith(".shp") or name.endswith(".mdb"):
								dataStructure.append(os.path.join(root, singleFile).replace(os.path.join(rootDir, single), ""))
					return dataStructure

	def check_data_structure(self):
		dataStructure = self.dataStructure[:]
		for i in dataStructure:
			checkPath = self.path + i
			check = False
			if os.path.isfile(checkPath):
				check = True
			_type = (os.path.basename(checkPath)).split('.')[-1]
			yield {"check": check, "type": _type, "path": checkPath}


class ShapefileValidator(object):
	def __init__(self, shapefile, crs=32633):
		super(ShapefileValidator, self).__init__()

		self.__name = os.path.basename(shapefile).split('.')[0]
		self.__shapefile = QgsVectorLayer(shapefile, 'shapefile', 'ogr')
		self.__crs = QgsCoordinateReferenceSystem(crs)

		d = shelve.open(os.path.join(os.getcwd(),"cfg","v4_data"))
		self.__fieldsStructure = d["shpFields"][self.name]
		if self.name == 'Instab':
			stab_list = d["shpFields"]['Stab'][3]['domain']
			instab_list = d["shpFields"]['Instab'][3]['domain']
			new_instab_list = instab_list[:]
			for x in instab_list:
				for y in stab_list:
					if y > 2000:
						new_instab_list.append(int("%d%d" % (x,y)))
			self.__fieldsStructure[3]['domain'] = new_instab_list
		d.close()

	@property
	def name(self):
		return self.__name

	@property
	def crs(self):
		return self.__crs

	def setCrs(self, value):
		self.__crs = QgsCoordinateReferenceSystem(value)

	@property
	def shapefile(self):
		return self.__shapefile

	@property
	def isValid(self):
		if self.shapefile.isValid():
			return True
		return False

	@property
	def hasFeatures(self):
		if self.shapefile.featureCount():
			return True
		return False

	def getFeatures(self, request=None):
		if request is None:
			return [feature for feature in self.shapefile.getFeatures()]
		return [feature for feature in self.shapefile.getFeatures(request)]

	def getGeometries(self):
		return [feature.geometry() for feature in self.getFeatures()]

	@property
	def hasValidCrs(self):
		if self.shapefile.crs() == self.crs:
			return True
		return False

	@property
	def fields(self):
		return [field for field in self.shapefile.pendingFields()]

	def hasField(self, name):
		if name in [field.name() for field in self.fields]:
			return True
		return False


	@property
	def fieldsStructure(self):
		return self.__fieldsStructure


	def validateFields(self):
		actualFields = {field.name() for field in self.fields}
		minimumFields = {structure["name"] for structure in self.fieldsStructure}
		for field in minimumFields:
			if not field in actualFields:
				yield {"field": field, "minField": True, "check": False}
			else:
				for qfield in self.fields:
					if qfield.name() == field:
						validator = FieldValidator(qfield, self.fieldsStructure)
						yield validator.check()
		"""
		for field in self.fields:
			validator = FieldValidator(field, self.fieldsStructure)
			yield validator.check()
		"""

	def validateGeometries(self):
		for feature in self.getFeatures():
			validator = GeometryValidator(feature, self.name)
			yield validator.check()

	def validateTopology(self, other=None):
		if other is None:
			featuresCombo = combinations(self.getFeatures(), 2)
			for combo in featuresCombo:
				feature, otherFeature = combo
				validator = TopologyValidator(feature, otherFeature,self.name,False)
				yield validator.check()
		else:
			for feature in self.getFeatures():
				for otherFeature in other.getFeatures():
					validator = TopologyValidator(feature, otherFeature,self.name,True)
					yield validator.check()

	def validateAttributes(self, fieldsList=None):
		if fieldsList is None:
			#fieldsList = self.fields
			actualFields = {field.name() for field in self.fields} #aggiunto
			minimumFields = {structure["name"] for structure in self.fieldsStructure} #aggiunto
			for field in minimumFields: #aggiunto
				if not field in actualFields: #aggiunto
					yield {"field": field, "minField": True, "check": False} #aggiunto
				else: #aggiunto
					for qfield in self.fields: #aggiunto
						if qfield.name() == field: #aggiunto
							validator = AttributeValidator(qfield, self.fieldsStructure, self.fields, self.shapefile, self.name) #aggiunto
							if validator.hasDomain(): #aggiunto
								yield validator.check() #aggiunto
		#for field in fieldsList:
		   # validator = AttributeValidator(field, self.fieldsStructure, self.fields, self.shapefile)
		   # if validator.hasDomain():
			 #   yield validator.check()




class FieldValidator(object):

	def __init__(self, field, fieldsStructure):
		super(FieldValidator, self).__init__()

		self.__field = field
		self.__fieldsStructure = fieldsStructure

	@property
	def fieldName(self):
		return self.__field.name()

	@property
	def fieldType(self):
		return self.__field.typeName()

	@property
	def fieldLength(self):
		return self.__field.length()

	@property
	def fieldPrecision(self):
		return self.__field.precision()

	@property
	def fieldsStructure(self):
		return self.__fieldsStructure

	def getStructure(self):
		for structure in self.fieldsStructure:
			if structure["name"] == self.fieldName:
				return structure

	def check(self):
		minimumFields = {structure["name"] for structure in self.fieldsStructure}

		properties = {}
		structure = self.getStructure()
		if structure:
			properties["name"] = False if self.fieldName != structure["name"] else True
			properties["type"] = False if self.fieldType != structure["type"] else True
			#properties["length"] = False if self.fieldLength != structure["length"] else True
			#properties["precision"] = False if self.fieldPrecision != structure["precision"] else True
		else:
			properties["name"] = False
			expectedName = {"name": [field["name"] for field in self.fieldsStructure]}
		check = all(properties.values())
		return {"field": self.fieldName, "check": check, "properties": properties, "expected": structure if properties["name"] else expectedName}


class GeometryValidator(object):

	def __init__(self, feature,name):
		super(GeometryValidator, self).__init__()

		self.__geometry = feature.geometry()
		self.__id = feature.id()
		self.__name = name
		self.feature = feature

	@property
	def geometry(self):
		return self.__geometry

	@property
	def id(self):
		return self.__id

	def check(self):
		check = True
		if self.geometry is None: #modificato Gianluca
			check = False #modificato Gianluca
			errs = ("La feature con id " + str(self.getFeatId()) + " non ha geometria",) #modificato Gianluca
		else:	 #modificato Gianluca
			errs = self.geometry.validateGeometry()
			if errs:
				check = False
				new_err = []
				for err in errs:
					if err.hasWhere():
						errText = err.what() + " (feature con id: " + str(self.getFeatId()) + ")"
						new_err.append((errText,err.where()))
				errs = new_err		
		return {"check": check, "errors": errs}


	@property
	def campoId(self):
		return DICT_IDS[self.name]
	def getFeatId(self):
		return self.feature[self.campoId]
	@property
	def name(self):
		return self.__name

class TopologyValidator(object):

	def __init__(self, feature, otherFeature,name, other):
		super(TopologyValidator, self).__init__()

		self.__geometry = feature.geometry()
		self.__otherGeometry = otherFeature.geometry()
		self.feature = feature
		self.otherFeature = otherFeature
		self.nameTable = name
		self.other = other

	@property
	def geometry(self):
		return self.__geometry

	@property
	def otherGeometry(self):
		return self.__otherGeometry

	def isOverlapped(self):
		return self.geometry.intersects(self.otherGeometry)

	def overlappedGeometry(self):
		return self.geometry.intersection(self.otherGeometry)

	def isGeometryValid(self):
		if self.geometry is None or not self.geometry.isGeosValid():
			return False
		return True

	def isOtherGeometryValid(self):
		if self.otherGeometry is None or not self.otherGeometry.isGeosValid():
			return False
		return True

	def check(self):
		info = ''
		check = True
		intersection = None
		if self.isGeometryValid() and self.isOtherGeometryValid():
			if self.isOverlapped():
				intersection = self.overlappedGeometry()
				if intersection.type() <= 1 or intersection.area() < 10 or (intersection.length() > 0 and (4 * math.pi * intersection.area() / intersection.length()**2) < 0.01):
					intersection = None
				else:
					check = False
					if self.other == False:
						info = "Dato: " + self.nameTable + "; Features:" + "{} - {}".format(self.feature[self.campoId], self.otherFeature[self.campoId])
					else:
						info = "Dato: " + self.nameTable + "; Feature:" + "{}".format(self.feature[self.campoId])					
					#try:
					#	QMessageBox.information(None, "Info", "{} - {}".format(self.feature["Id_gt"], self.otherFeature["Id_gt"]))
					#except:
					#	pass
		return {"check": check, "intersection": intersection, "info": info}

	@property
	def campoId(self):
		return DICT_IDS[self.nameTable]

class AttributeValidator(object):

	def __init__(self, field, fieldsStructure, fieldsList, shapefile,name):
		super(AttributeValidator, self).__init__()

		self.__field = field
		self.__fieldsList = fieldsList
		self.__shapefile = shapefile
		self.__name = name
		for field in fieldsStructure:
			if field["name"] == self.fieldName:
				self.__fieldStructure = field

	@property
	def shapefile(self):
		return self.__shapefile

	@property
	def field(self):
		return self.__field

	@property
	def fields(self):
		return self.__fieldsList

	@property
	def fieldName(self):
		return self.field.name()

	@property
	def fieldStructure(self):
		return self.__fieldStructure

	def hasDomain(self):
		if "domain" in self.fieldStructure.keys():
			return True
		return False

	@property
	def domain(self):
		if self.hasDomain():
			return self.fieldStructure["domain"]

	def check(self):
		check = False
		errs = []
		#request = QgsFeatureRequest()
		#request.setSubsetOfAttributes([self.fieldName], self.shapefile.pendingFields())
		#request.setFlags(QgsFeatureRequest.NoGeometry)
		domain = [str(i) for i in self.domain]
		for row in self.shapefile.getFeatures():
			if (str(row[self.fieldName]) == 'NULL' or str(row[self.fieldName]) == '0'):
				continue
			if str(row[self.fieldName]) not in domain:
				errs.append(row[self.campoId])
		if not errs:
			check = True
		return {"field": self.fieldName, "check": check, "featuresId": errs}

	@property
	def campoId(self):
		return DICT_IDS[self.name]

	@property
	def name(self):
		return self.__name